# pylint: disable=line-too-long
"""The histogram commands module.

These commands are used in the following models:
LPD6, MSO4, MSO4B, MSO5, MSO5B, MSO5LP, MSO6, MSO6B

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:
    ```
    - HISTogram:ADDNew <QString>
    - HISTogram:DELETEALL
    - HISTogram:DELete <QString>
    - HISTogram:HISTogram<x>:BOX <NR3>,<NR3>,<NR3>,<NR3>
    - HISTogram:HISTogram<x>:BOX?
    - HISTogram:HISTogram<x>:BSTate {ON|OFF}
    - HISTogram:HISTogram<x>:BSTate?
    - HISTogram:HISTogram<x>:DATa?
    - HISTogram:HISTogram<x>:DISPlay {LINEAr|LOG}
    - HISTogram:HISTogram<x>:DISPlay?
    - HISTogram:HISTogram<x>:FUNCtion {HORizontal|VERTical}
    - HISTogram:HISTogram<x>:FUNCtion?
    - HISTogram:HISTogram<x>:MEASurement:COUNt {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:COUNt?
    - HISTogram:HISTogram<x>:MEASurement:HITS {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:HITS?
    - HISTogram:HISTogram<x>:MEASurement:MAX {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:MAX?
    - HISTogram:HISTogram<x>:MEASurement:MEAN {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:MEAN?
    - HISTogram:HISTogram<x>:MEASurement:MEDian {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:MEDian?
    - HISTogram:HISTogram<x>:MEASurement:MIN {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:MIN?
    - HISTogram:HISTogram<x>:MEASurement:MODE {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:MODE?
    - HISTogram:HISTogram<x>:MEASurement:ONESigma {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:ONESigma?
    - HISTogram:HISTogram<x>:MEASurement:PHITs {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:PHITs?
    - HISTogram:HISTogram<x>:MEASurement:PK2PK {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:PK2PK?
    - HISTogram:HISTogram<x>:MEASurement:RESUlts? {COUNt|HITS|MAX|MIN|PK2PK|MODE|MEAN|MEDian|PHITs|STDDev|ONESigma|TWOSigma|THRSigma,ALLAcqs|CURRentacq|HISTory,MAXimum|MEAN|MINimum|PK2PK|POPUlation|STDDev}
    - HISTogram:HISTogram<x>:MEASurement:STDDev {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:STDDev?
    - HISTogram:HISTogram<x>:MEASurement:THRSigma {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:THRSigma?
    - HISTogram:HISTogram<x>:MEASurement:TWOSigma {ON|OFF}
    - HISTogram:HISTogram<x>:MEASurement:TWOSigma?
    - HISTogram:HISTogram<x>:SAVe <QString>
    - HISTogram:HISTogram<x>:SIZe <NR3>
    - HISTogram:HISTogram<x>:SIZe?
    - HISTogram:HISTogram<x>:SOUrce {CH<x>|MATH<x>|REF<x>}
    - HISTogram:HISTogram<x>:SOUrce?
    - HISTogram:HISTogram<x>:STATE {ON|OFF}
    - HISTogram:HISTogram<x>:STATE?
    - HISTogram:HISTogram<x>:TRANsparency <NR3>
    - HISTogram:HISTogram<x>:TRANsparency?
    - HISTogram:LIST?
    ```
"""  # noqa: E501

from typing import Dict, Optional, TYPE_CHECKING

from ..helpers import (
    DefaultDictPassKeyToFactory,
    SCPICmdRead,
    SCPICmdReadWithArguments,
    SCPICmdWrite,
    SCPICmdWriteNoArguments,
    ValidatedDynamicNumberCmd,
)

if TYPE_CHECKING:
    from tm_devices.driver_mixins.device_control.pi_control import PIControl


class HistogramList(SCPICmdRead):
    """The ``HISTogram:LIST`` command.

    Description:
        - This query returns a comma separated list of all currently defined histograms.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:LIST?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:LIST?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - HISTogram:LIST?
        ```
    """


class HistogramHistogramItemTransparency(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:TRANsparency`` command.

    Description:
        - This command sets or queries the transparency of the histogram bins.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:TRANsparency?``
          query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:TRANsparency?``
          query and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:TRANsparency value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:TRANsparency <NR3>
        - HISTogram:HISTogram<x>:TRANsparency?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``<NR3>`` specifies the transparency as a percentage, with a minimum of 0 and a maximum of
          100.
    """


class HistogramHistogramItemState(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:STATE`` command.

    Description:
        - This command sets or queries whether histogram calculations are enabled for the specified
          histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:STATE?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:STATE?`` query
          and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:STATE value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:STATE {ON|OFF}
        - HISTogram:HISTogram<x>:STATE?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the histogram calculations.
        - ``OFF`` disables the histogram calculations.
    """


class HistogramHistogramItemSource(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:SOUrce`` command.

    Description:
        - This command sets or queries which source waveform will be compared against the histogram
          box when the histogram testing is enabled.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:SOUrce?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:SOUrce?`` query
          and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:SOUrce value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:SOUrce {CH<x>|MATH<x>|REF<x>}
        - HISTogram:HISTogram<x>:SOUrce?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``CH<x>`` specifies an analog channel as source.
        - ``MATH<x>`` specifies a math channel as source.
        - ``REF<x>`` specifies a reference waveform as the source.
    """


class HistogramHistogramItemSize(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:SIZe`` command.

    Description:
        - This command sets or queries the height or width of the specified histogram bins in
          divisions. This size can be larger than the box, which allows you to view the histogram
          bins easily.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:SIZe?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:SIZe?`` query
          and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:SIZe value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:SIZe <NR3>
        - HISTogram:HISTogram<x>:SIZe?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``<NR3>`` specifies the number of divisions to set the height or width of the histogram
          bins to.
    """


class HistogramHistogramItemSave(SCPICmdWrite):
    """The ``HISTogram:HISTogram<x>:SAVe`` command.

    Description:
        - This command saves the specified histograms data as a comma separated list of values.

    Usage:
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:SAVe value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:SAVe <QString>
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``<QString>`` is the file path to save the.csv file to.
    """

    _WRAP_ARG_WITH_QUOTES = True


class HistogramHistogramItemMeasurementTwosigma(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:TWOSigma`` command.

    Description:
        - This command sets or queries whether the μ±2(sigma) measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:TWOSigma?`` query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:TWOSigma?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:TWOSigma value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:TWOSigma {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:TWOSigma?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the μ±2(sigma) measurement.
        - ``OFF`` disables the μ±2(sigma) measurement.
    """


class HistogramHistogramItemMeasurementThrsigma(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:THRSigma`` command.

    Description:
        - This command sets or queries whether the μ±3(sigma) measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:THRSigma?`` query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:THRSigma?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:THRSigma value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:THRSigma {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:THRSigma?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the μ±3(sigma) measurement.
        - ``OFF`` disables the μ±3(sigma) measurement.
    """


class HistogramHistogramItemMeasurementStddev(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:STDDev`` command.

    Description:
        - This command sets or queries whether the Standard Deviation measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:STDDev?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:STDDev?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:STDDev value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:STDDev {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:STDDev?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Standard Deviation measurement.
        - ``OFF`` disables the Standard Deviation measurement.
    """


class HistogramHistogramItemMeasurementResults(SCPICmdReadWithArguments):
    """The ``HISTogram:HISTogram<x>:MEASurement:RESUlts`` command.

    Description:
        - This query only command returns the measurement results from the specified histogram. The
          argument must be in the form of {argument 1},{argument 2},{argument 3}. Argument 1
          specifies which histogram measurement to return results for. Argument 2 specifies which
          acquisitions to return measurement results for. Argument 3 specifies which statistic to
          return measurement results for.

    Usage:
        - Using the ``.query(argument)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:RESUlts? argument`` query.
        - Using the ``.verify(argument, value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:RESUlts? argument`` query and raise an AssertionError
          if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:RESUlts? {COUNt|HITS|MAX|MIN|PK2PK|MODE|MEAN|MEDian|PHITs|STDDev|ONESigma|TWOSigma|THRSigma,ALLAcqs|CURRentacq|HISTory,MAXimum|MEAN|MINimum|PK2PK|POPUlation|STDDev}
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``COUNt`` specifies the Count as the histogram measurement to return results for.
        - ``HITS`` specifies the Hits as the histogram measurement to return results for.
        - ``MAX`` specifies the Max as the histogram measurement to return results for.
        - ``MIN`` specifies the Min as the histogram measurement to return results for.
        - ``PK2PK`` specifies the Peak-to-peak as the histogram measurement or statistic to return
          results for.
        - ``MODE`` specifies the Mode as the histogrammeasurement to return results for.
        - ``MEAN`` specifies the Mean as the histogram measurement or statistic to return results
          for.
        - ``MEDian`` specifies the Median as the histogram measurement to return results for.
        - ``PHITs`` specifies the Peak Hits as the histogram measurement to return results for.
        - ``STDDev`` specifies the Standard Deviation as the histogram measurement or statistic to
          return results for.
        - ``ONESigma`` specifies the μ±1(sigma) as the histogram measurement to return results for.
        - ``TWOSigma`` specifies the μ±2(sigma) as the histogram measurement to return results for.
        - ``THRSigma`` specifies the μ±3(sigma) as the histogram measurement to return results for.
        - ``ALLAcqs`` specifies the All Acquisitions as the acquisitions to return results for.
        - ``CURRentacq`` specifies the Current Acquisitions as the acquisitions to return results
          for.
        - ``HISTory`` specifies the History as the acquisitions to return results for.
        - ``MAXimum`` specifies the Maximum as the statistic to return results for.
        - ``MINimum`` specifies the Minimum as the statistic to return results for.
        - ``POPUlation`` specifies the Population as the statistic to return results for.
    """  # noqa: E501


class HistogramHistogramItemMeasurementPk2pk(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:PK2PK`` command.

    Description:
        - This command sets or queries whether the Peak-to-peak measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:PK2PK?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:PK2PK?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:PK2PK value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:PK2PK {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:PK2PK?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Peak-to-peak measurement.
        - ``OFF`` disables the Peak-to-peak measurement.
    """


class HistogramHistogramItemMeasurementPhits(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:PHITs`` command.

    Description:
        - This command sets or queries whether the Peak Hits measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:PHITs?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:PHITs?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:PHITs value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:PHITs {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:PHITs?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Peak Hits measurement.
        - ``OFF`` disables the Peak Hits measurement.
    """


class HistogramHistogramItemMeasurementOnesigma(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:ONESigma`` command.

    Description:
        - This command sets or queries whether the μ±1(sigma) measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:ONESigma?`` query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:ONESigma?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:ONESigma value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:ONESigma {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:ONESigma?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the μ±1(sigma) measurement.
        - ``OFF`` disables the μ±1(sigma) measurement.
    """


class HistogramHistogramItemMeasurementMode(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:MODE`` command.

    Description:
        - This command sets or queries whether the Mode measurement is enabled on the histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:MODE?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MODE?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MODE value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:MODE {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:MODE?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``HORizontal`` sets the histogram mode to horizontal.
        - ``VERTical`` sets the histogram mode to vertical.
    """


class HistogramHistogramItemMeasurementMin(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:MIN`` command.

    Description:
        - This command sets or queries whether the Min measurement is enabled on the histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:MIN?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MIN?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MIN value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:MIN {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:MIN?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Min measurement.
        - ``OFF`` disables the Min measurement.
    """


class HistogramHistogramItemMeasurementMedian(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:MEDian`` command.

    Description:
        - This command sets or queries whether the Median measurement is enabled on the histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:MEDian?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MEDian?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MEDian value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:MEDian {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:MEDian?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Median measurement.
        - ``OFF`` disables the Median measurement.
    """


class HistogramHistogramItemMeasurementMean(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:MEAN`` command.

    Description:
        - This command sets or queries whether the Mean measurement is enabled on the histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:MEAN?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MEAN?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MEAN value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:MEAN {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:MEAN?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Mean measurement.
        - ``OFF`` disables the Mean measurement.
    """


class HistogramHistogramItemMeasurementMax(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:MAX`` command.

    Description:
        - This command sets or queries whether the Max measurement is enabled on the histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:MAX?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MAX?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:MAX value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:MAX {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:MAX?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Max measurement.
        - ``OFF`` disables the Max measurement.
    """


class HistogramHistogramItemMeasurementHits(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:HITS`` command.

    Description:
        - This command sets or queries whether the Hits in Box measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:HITS?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:HITS?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:HITS value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:HITS {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:HITS?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the Hits in Box measurement.
        - ``OFF`` disables the Hits in Box measurement.
    """


class HistogramHistogramItemMeasurementCount(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement:COUNt`` command.

    Description:
        - This command sets or queries whether the waveform count measurement is enabled on the
          histogram.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement:COUNt?``
          query.
        - Using the ``.verify(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:COUNt?`` query and raise an AssertionError if the
          returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the
          ``HISTogram:HISTogram<x>:MEASurement:COUNt value`` command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:MEASurement:COUNt {ON|OFF}
        - HISTogram:HISTogram<x>:MEASurement:COUNt?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the waveform count measurement.
        - ``OFF`` disables the waveform count measurement.
    """


#  pylint: disable=too-many-instance-attributes
class HistogramHistogramItemMeasurement(SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:MEASurement`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:MEASurement?``
          query and raise an AssertionError if the returned value does not match ``value``.

    Info:
        - ``HISTogram<x>`` specifies the histogram number.

    Properties:
        - ``.count``: The ``HISTogram:HISTogram<x>:MEASurement:COUNt`` command.
        - ``.hits``: The ``HISTogram:HISTogram<x>:MEASurement:HITS`` command.
        - ``.max``: The ``HISTogram:HISTogram<x>:MEASurement:MAX`` command.
        - ``.mean``: The ``HISTogram:HISTogram<x>:MEASurement:MEAN`` command.
        - ``.median``: The ``HISTogram:HISTogram<x>:MEASurement:MEDian`` command.
        - ``.min``: The ``HISTogram:HISTogram<x>:MEASurement:MIN`` command.
        - ``.mode``: The ``HISTogram:HISTogram<x>:MEASurement:MODE`` command.
        - ``.onesigma``: The ``HISTogram:HISTogram<x>:MEASurement:ONESigma`` command.
        - ``.phits``: The ``HISTogram:HISTogram<x>:MEASurement:PHITs`` command.
        - ``.pk2pk``: The ``HISTogram:HISTogram<x>:MEASurement:PK2PK`` command.
        - ``.results``: The ``HISTogram:HISTogram<x>:MEASurement:RESUlts`` command.
        - ``.stddev``: The ``HISTogram:HISTogram<x>:MEASurement:STDDev`` command.
        - ``.thrsigma``: The ``HISTogram:HISTogram<x>:MEASurement:THRSigma`` command.
        - ``.twosigma``: The ``HISTogram:HISTogram<x>:MEASurement:TWOSigma`` command.
    """

    def __init__(self, device: Optional["PIControl"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._count = HistogramHistogramItemMeasurementCount(device, f"{self._cmd_syntax}:COUNt")
        self._hits = HistogramHistogramItemMeasurementHits(device, f"{self._cmd_syntax}:HITS")
        self._max = HistogramHistogramItemMeasurementMax(device, f"{self._cmd_syntax}:MAX")
        self._mean = HistogramHistogramItemMeasurementMean(device, f"{self._cmd_syntax}:MEAN")
        self._median = HistogramHistogramItemMeasurementMedian(device, f"{self._cmd_syntax}:MEDian")
        self._min = HistogramHistogramItemMeasurementMin(device, f"{self._cmd_syntax}:MIN")
        self._mode = HistogramHistogramItemMeasurementMode(device, f"{self._cmd_syntax}:MODE")
        self._onesigma = HistogramHistogramItemMeasurementOnesigma(
            device, f"{self._cmd_syntax}:ONESigma"
        )
        self._phits = HistogramHistogramItemMeasurementPhits(device, f"{self._cmd_syntax}:PHITs")
        self._pk2pk = HistogramHistogramItemMeasurementPk2pk(device, f"{self._cmd_syntax}:PK2PK")
        self._results = HistogramHistogramItemMeasurementResults(
            device, f"{self._cmd_syntax}:RESUlts"
        )
        self._stddev = HistogramHistogramItemMeasurementStddev(device, f"{self._cmd_syntax}:STDDev")
        self._thrsigma = HistogramHistogramItemMeasurementThrsigma(
            device, f"{self._cmd_syntax}:THRSigma"
        )
        self._twosigma = HistogramHistogramItemMeasurementTwosigma(
            device, f"{self._cmd_syntax}:TWOSigma"
        )

    @property
    def count(self) -> HistogramHistogramItemMeasurementCount:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:COUNt`` command.

        Description:
            - This command sets or queries whether the waveform count measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:COUNt?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:COUNt?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:COUNt value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:COUNt {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:COUNt?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the waveform count measurement.
            - ``OFF`` disables the waveform count measurement.
        """
        return self._count

    @property
    def hits(self) -> HistogramHistogramItemMeasurementHits:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:HITS`` command.

        Description:
            - This command sets or queries whether the Hits in Box measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:HITS?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:HITS?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:HITS value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:HITS {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:HITS?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Hits in Box measurement.
            - ``OFF`` disables the Hits in Box measurement.
        """
        return self._hits

    @property
    def max(self) -> HistogramHistogramItemMeasurementMax:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:MAX`` command.

        Description:
            - This command sets or queries whether the Max measurement is enabled on the histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MAX?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MAX?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MAX value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:MAX {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:MAX?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Max measurement.
            - ``OFF`` disables the Max measurement.
        """
        return self._max

    @property
    def mean(self) -> HistogramHistogramItemMeasurementMean:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:MEAN`` command.

        Description:
            - This command sets or queries whether the Mean measurement is enabled on the histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MEAN?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MEAN?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MEAN value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:MEAN {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:MEAN?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Mean measurement.
            - ``OFF`` disables the Mean measurement.
        """
        return self._mean

    @property
    def median(self) -> HistogramHistogramItemMeasurementMedian:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:MEDian`` command.

        Description:
            - This command sets or queries whether the Median measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MEDian?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MEDian?`` query and raise an AssertionError if
              the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MEDian value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:MEDian {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:MEDian?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Median measurement.
            - ``OFF`` disables the Median measurement.
        """
        return self._median

    @property
    def min(self) -> HistogramHistogramItemMeasurementMin:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:MIN`` command.

        Description:
            - This command sets or queries whether the Min measurement is enabled on the histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MIN?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MIN?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MIN value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:MIN {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:MIN?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Min measurement.
            - ``OFF`` disables the Min measurement.
        """
        return self._min

    @property
    def mode(self) -> HistogramHistogramItemMeasurementMode:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:MODE`` command.

        Description:
            - This command sets or queries whether the Mode measurement is enabled on the histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MODE?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MODE?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:MODE value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:MODE {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:MODE?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``HORizontal`` sets the histogram mode to horizontal.
            - ``VERTical`` sets the histogram mode to vertical.
        """
        return self._mode

    @property
    def onesigma(self) -> HistogramHistogramItemMeasurementOnesigma:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:ONESigma`` command.

        Description:
            - This command sets or queries whether the μ±1(sigma) measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:ONESigma?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:ONESigma?`` query and raise an AssertionError if
              the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:ONESigma value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:ONESigma {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:ONESigma?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the μ±1(sigma) measurement.
            - ``OFF`` disables the μ±1(sigma) measurement.
        """
        return self._onesigma

    @property
    def phits(self) -> HistogramHistogramItemMeasurementPhits:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:PHITs`` command.

        Description:
            - This command sets or queries whether the Peak Hits measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:PHITs?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:PHITs?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:PHITs value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:PHITs {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:PHITs?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Peak Hits measurement.
            - ``OFF`` disables the Peak Hits measurement.
        """
        return self._phits

    @property
    def pk2pk(self) -> HistogramHistogramItemMeasurementPk2pk:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:PK2PK`` command.

        Description:
            - This command sets or queries whether the Peak-to-peak measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:PK2PK?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:PK2PK?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:PK2PK value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:PK2PK {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:PK2PK?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Peak-to-peak measurement.
            - ``OFF`` disables the Peak-to-peak measurement.
        """
        return self._pk2pk

    @property
    def results(self) -> HistogramHistogramItemMeasurementResults:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:RESUlts`` command.

        Description:
            - This query only command returns the measurement results from the specified histogram.
              The argument must be in the form of {argument 1},{argument 2},{argument 3}. Argument 1
              specifies which histogram measurement to return results for. Argument 2 specifies
              which acquisitions to return measurement results for. Argument 3 specifies which
              statistic to return measurement results for.

        Usage:
            - Using the ``.query(argument)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:RESUlts? argument`` query.
            - Using the ``.verify(argument, value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:RESUlts? argument`` query and raise an
              AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:RESUlts? {COUNt|HITS|MAX|MIN|PK2PK|MODE|MEAN|MEDian|PHITs|STDDev|ONESigma|TWOSigma|THRSigma,ALLAcqs|CURRentacq|HISTory,MAXimum|MEAN|MINimum|PK2PK|POPUlation|STDDev}
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``COUNt`` specifies the Count as the histogram measurement to return results for.
            - ``HITS`` specifies the Hits as the histogram measurement to return results for.
            - ``MAX`` specifies the Max as the histogram measurement to return results for.
            - ``MIN`` specifies the Min as the histogram measurement to return results for.
            - ``PK2PK`` specifies the Peak-to-peak as the histogram measurement or statistic to
              return results for.
            - ``MODE`` specifies the Mode as the histogrammeasurement to return results for.
            - ``MEAN`` specifies the Mean as the histogram measurement or statistic to return
              results for.
            - ``MEDian`` specifies the Median as the histogram measurement to return results for.
            - ``PHITs`` specifies the Peak Hits as the histogram measurement to return results for.
            - ``STDDev`` specifies the Standard Deviation as the histogram measurement or statistic
              to return results for.
            - ``ONESigma`` specifies the μ±1(sigma) as the histogram measurement to return results
              for.
            - ``TWOSigma`` specifies the μ±2(sigma) as the histogram measurement to return results
              for.
            - ``THRSigma`` specifies the μ±3(sigma) as the histogram measurement to return results
              for.
            - ``ALLAcqs`` specifies the All Acquisitions as the acquisitions to return results for.
            - ``CURRentacq`` specifies the Current Acquisitions as the acquisitions to return
              results for.
            - ``HISTory`` specifies the History as the acquisitions to return results for.
            - ``MAXimum`` specifies the Maximum as the statistic to return results for.
            - ``MINimum`` specifies the Minimum as the statistic to return results for.
            - ``POPUlation`` specifies the Population as the statistic to return results for.
        """  # noqa: E501
        return self._results

    @property
    def stddev(self) -> HistogramHistogramItemMeasurementStddev:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:STDDev`` command.

        Description:
            - This command sets or queries whether the Standard Deviation measurement is enabled on
              the histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:STDDev?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:STDDev?`` query and raise an AssertionError if
              the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:STDDev value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:STDDev {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:STDDev?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the Standard Deviation measurement.
            - ``OFF`` disables the Standard Deviation measurement.
        """
        return self._stddev

    @property
    def thrsigma(self) -> HistogramHistogramItemMeasurementThrsigma:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:THRSigma`` command.

        Description:
            - This command sets or queries whether the μ±3(sigma) measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:THRSigma?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:THRSigma?`` query and raise an AssertionError if
              the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:THRSigma value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:THRSigma {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:THRSigma?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the μ±3(sigma) measurement.
            - ``OFF`` disables the μ±3(sigma) measurement.
        """
        return self._thrsigma

    @property
    def twosigma(self) -> HistogramHistogramItemMeasurementTwosigma:
        """Return the ``HISTogram:HISTogram<x>:MEASurement:TWOSigma`` command.

        Description:
            - This command sets or queries whether the μ±2(sigma) measurement is enabled on the
              histogram.

        Usage:
            - Using the ``.query()`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:TWOSigma?`` query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:TWOSigma?`` query and raise an AssertionError if
              the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement:TWOSigma value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:MEASurement:TWOSigma {ON|OFF}
            - HISTogram:HISTogram<x>:MEASurement:TWOSigma?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the μ±2(sigma) measurement.
            - ``OFF`` disables the μ±2(sigma) measurement.
        """
        return self._twosigma


class HistogramHistogramItemFunction(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:FUNCtion`` command.

    Description:
        - This command sets or queries the histogram mode.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:FUNCtion?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:FUNCtion?``
          query and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:FUNCtion value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:FUNCtion {HORizontal|VERTical}
        - HISTogram:HISTogram<x>:FUNCtion?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``HORizontal`` sets the histogram mode to horizontal.
        - ``VERTical`` sets the histogram mode to vertical.
    """


class HistogramHistogramItemDisplay(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:DISPlay`` command.

    Description:
        - This command sets or queries the histogram scaling display setting.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:DISPlay?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:DISPlay?``
          query and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:DISPlay value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:DISPlay {LINEAr|LOG}
        - HISTogram:HISTogram<x>:DISPlay?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``LINEAr`` sets the histogram scaling to linear mode.
        - ``LOG`` sets the histogram scaling to logarithmic mode.
    """


class HistogramHistogramItemData(SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:DATa`` command.

    Description:
        - This query only command returns a comma separated list of numbers representing the
          histogram bin data.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:DATa?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:DATa?`` query
          and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:DATa?
        ```
    """


class HistogramHistogramItemBstate(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:BSTate`` command.

    Description:
        - This command sets or queries whether the histogram badge is displayed.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:BSTate?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:BSTate?`` query
          and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:BSTate value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:BSTate {ON|OFF}
        - HISTogram:HISTogram<x>:BSTate?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``ON`` enables the histogram badge display.
        - ``OFF`` disables the histogram badge display.
    """


class HistogramHistogramItemBox(SCPICmdWrite, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>:BOX`` command.

    Description:
        - This command sets or queries the top, left, bottom and right positions of the histogram
          box in source waveform coordinates. The argument is of the form
          <top>,<left>,<bottom>,<right>.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:BOX?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:BOX?`` query
          and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:BOX value``
          command.

    SCPI Syntax:
        ```
        - HISTogram:HISTogram<x>:BOX <NR3>,<NR3>,<NR3>,<NR3>
        - HISTogram:HISTogram<x>:BOX?
        ```

    Info:
        - ``HISTogram<x>`` specifies the histogram number.
        - ``<NR3>`` specifies four position values, separated by commas. The values are the top,
          left, bottom, and right coordinates in that order.
    """


#  pylint: disable=too-many-instance-attributes
class HistogramHistogramItem(ValidatedDynamicNumberCmd, SCPICmdRead):
    """The ``HISTogram:HISTogram<x>`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    Info:
        - ``HISTogram<x>`` specifies the histogram number.

    Properties:
        - ``.box``: The ``HISTogram:HISTogram<x>:BOX`` command.
        - ``.bstate``: The ``HISTogram:HISTogram<x>:BSTate`` command.
        - ``.data``: The ``HISTogram:HISTogram<x>:DATa`` command.
        - ``.display``: The ``HISTogram:HISTogram<x>:DISPlay`` command.
        - ``.function``: The ``HISTogram:HISTogram<x>:FUNCtion`` command.
        - ``.measurement``: The ``HISTogram:HISTogram<x>:MEASurement`` command tree.
        - ``.save``: The ``HISTogram:HISTogram<x>:SAVe`` command.
        - ``.size``: The ``HISTogram:HISTogram<x>:SIZe`` command.
        - ``.source``: The ``HISTogram:HISTogram<x>:SOUrce`` command.
        - ``.state``: The ``HISTogram:HISTogram<x>:STATE`` command.
        - ``.transparency``: The ``HISTogram:HISTogram<x>:TRANsparency`` command.
    """

    def __init__(self, device: Optional["PIControl"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._box = HistogramHistogramItemBox(device, f"{self._cmd_syntax}:BOX")
        self._bstate = HistogramHistogramItemBstate(device, f"{self._cmd_syntax}:BSTate")
        self._data = HistogramHistogramItemData(device, f"{self._cmd_syntax}:DATa")
        self._display = HistogramHistogramItemDisplay(device, f"{self._cmd_syntax}:DISPlay")
        self._function = HistogramHistogramItemFunction(device, f"{self._cmd_syntax}:FUNCtion")
        self._measurement = HistogramHistogramItemMeasurement(
            device, f"{self._cmd_syntax}:MEASurement"
        )
        self._save = HistogramHistogramItemSave(device, f"{self._cmd_syntax}:SAVe")
        self._size = HistogramHistogramItemSize(device, f"{self._cmd_syntax}:SIZe")
        self._source = HistogramHistogramItemSource(device, f"{self._cmd_syntax}:SOUrce")
        self._state = HistogramHistogramItemState(device, f"{self._cmd_syntax}:STATE")
        self._transparency = HistogramHistogramItemTransparency(
            device, f"{self._cmd_syntax}:TRANsparency"
        )

    @property
    def box(self) -> HistogramHistogramItemBox:
        """Return the ``HISTogram:HISTogram<x>:BOX`` command.

        Description:
            - This command sets or queries the top, left, bottom and right positions of the
              histogram box in source waveform coordinates. The argument is of the form
              <top>,<left>,<bottom>,<right>.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:BOX?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:BOX?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:BOX value``
              command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:BOX <NR3>,<NR3>,<NR3>,<NR3>
            - HISTogram:HISTogram<x>:BOX?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``<NR3>`` specifies four position values, separated by commas. The values are the top,
              left, bottom, and right coordinates in that order.
        """
        return self._box

    @property
    def bstate(self) -> HistogramHistogramItemBstate:
        """Return the ``HISTogram:HISTogram<x>:BSTate`` command.

        Description:
            - This command sets or queries whether the histogram badge is displayed.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:BSTate?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:BSTate?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:BSTate value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:BSTate {ON|OFF}
            - HISTogram:HISTogram<x>:BSTate?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the histogram badge display.
            - ``OFF`` disables the histogram badge display.
        """
        return self._bstate

    @property
    def data(self) -> HistogramHistogramItemData:
        """Return the ``HISTogram:HISTogram<x>:DATa`` command.

        Description:
            - This query only command returns a comma separated list of numbers representing the
              histogram bin data.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:DATa?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:DATa?``
              query and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:DATa?
            ```
        """
        return self._data

    @property
    def display(self) -> HistogramHistogramItemDisplay:
        """Return the ``HISTogram:HISTogram<x>:DISPlay`` command.

        Description:
            - This command sets or queries the histogram scaling display setting.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:DISPlay?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:DISPlay?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:DISPlay value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:DISPlay {LINEAr|LOG}
            - HISTogram:HISTogram<x>:DISPlay?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``LINEAr`` sets the histogram scaling to linear mode.
            - ``LOG`` sets the histogram scaling to logarithmic mode.
        """
        return self._display

    @property
    def function(self) -> HistogramHistogramItemFunction:
        """Return the ``HISTogram:HISTogram<x>:FUNCtion`` command.

        Description:
            - This command sets or queries the histogram mode.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:FUNCtion?``
              query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:FUNCtion?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:FUNCtion value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:FUNCtion {HORizontal|VERTical}
            - HISTogram:HISTogram<x>:FUNCtion?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``HORizontal`` sets the histogram mode to horizontal.
            - ``VERTical`` sets the histogram mode to vertical.
        """
        return self._function

    @property
    def measurement(self) -> HistogramHistogramItemMeasurement:
        """Return the ``HISTogram:HISTogram<x>:MEASurement`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:MEASurement?``
              query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:MEASurement?`` query and raise an AssertionError if the
              returned value does not match ``value``.

        Info:
            - ``HISTogram<x>`` specifies the histogram number.

        Sub-properties:
            - ``.count``: The ``HISTogram:HISTogram<x>:MEASurement:COUNt`` command.
            - ``.hits``: The ``HISTogram:HISTogram<x>:MEASurement:HITS`` command.
            - ``.max``: The ``HISTogram:HISTogram<x>:MEASurement:MAX`` command.
            - ``.mean``: The ``HISTogram:HISTogram<x>:MEASurement:MEAN`` command.
            - ``.median``: The ``HISTogram:HISTogram<x>:MEASurement:MEDian`` command.
            - ``.min``: The ``HISTogram:HISTogram<x>:MEASurement:MIN`` command.
            - ``.mode``: The ``HISTogram:HISTogram<x>:MEASurement:MODE`` command.
            - ``.onesigma``: The ``HISTogram:HISTogram<x>:MEASurement:ONESigma`` command.
            - ``.phits``: The ``HISTogram:HISTogram<x>:MEASurement:PHITs`` command.
            - ``.pk2pk``: The ``HISTogram:HISTogram<x>:MEASurement:PK2PK`` command.
            - ``.results``: The ``HISTogram:HISTogram<x>:MEASurement:RESUlts`` command.
            - ``.stddev``: The ``HISTogram:HISTogram<x>:MEASurement:STDDev`` command.
            - ``.thrsigma``: The ``HISTogram:HISTogram<x>:MEASurement:THRSigma`` command.
            - ``.twosigma``: The ``HISTogram:HISTogram<x>:MEASurement:TWOSigma`` command.
        """
        return self._measurement

    @property
    def save(self) -> HistogramHistogramItemSave:
        """Return the ``HISTogram:HISTogram<x>:SAVe`` command.

        Description:
            - This command saves the specified histograms data as a comma separated list of values.

        Usage:
            - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:SAVe value``
              command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:SAVe <QString>
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``<QString>`` is the file path to save the.csv file to.
        """
        return self._save

    @property
    def size(self) -> HistogramHistogramItemSize:
        """Return the ``HISTogram:HISTogram<x>:SIZe`` command.

        Description:
            - This command sets or queries the height or width of the specified histogram bins in
              divisions. This size can be larger than the box, which allows you to view the
              histogram bins easily.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:SIZe?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:SIZe?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``HISTogram:HISTogram<x>:SIZe value``
              command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:SIZe <NR3>
            - HISTogram:HISTogram<x>:SIZe?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``<NR3>`` specifies the number of divisions to set the height or width of the
              histogram bins to.
        """
        return self._size

    @property
    def source(self) -> HistogramHistogramItemSource:
        """Return the ``HISTogram:HISTogram<x>:SOUrce`` command.

        Description:
            - This command sets or queries which source waveform will be compared against the
              histogram box when the histogram testing is enabled.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:SOUrce?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:SOUrce?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:SOUrce value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:SOUrce {CH<x>|MATH<x>|REF<x>}
            - HISTogram:HISTogram<x>:SOUrce?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``CH<x>`` specifies an analog channel as source.
            - ``MATH<x>`` specifies a math channel as source.
            - ``REF<x>`` specifies a reference waveform as the source.
        """
        return self._source

    @property
    def state(self) -> HistogramHistogramItemState:
        """Return the ``HISTogram:HISTogram<x>:STATE`` command.

        Description:
            - This command sets or queries whether histogram calculations are enabled for the
              specified histogram.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:STATE?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>:STATE?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:STATE value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:STATE {ON|OFF}
            - HISTogram:HISTogram<x>:STATE?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``ON`` enables the histogram calculations.
            - ``OFF`` disables the histogram calculations.
        """
        return self._state

    @property
    def transparency(self) -> HistogramHistogramItemTransparency:
        """Return the ``HISTogram:HISTogram<x>:TRANsparency`` command.

        Description:
            - This command sets or queries the transparency of the histogram bins.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>:TRANsparency?``
              query.
            - Using the ``.verify(value)`` method will send the
              ``HISTogram:HISTogram<x>:TRANsparency?`` query and raise an AssertionError if the
              returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``HISTogram:HISTogram<x>:TRANsparency value`` command.

        SCPI Syntax:
            ```
            - HISTogram:HISTogram<x>:TRANsparency <NR3>
            - HISTogram:HISTogram<x>:TRANsparency?
            ```

        Info:
            - ``HISTogram<x>`` specifies the histogram number.
            - ``<NR3>`` specifies the transparency as a percentage, with a minimum of 0 and a
              maximum of 100.
        """
        return self._transparency


class HistogramDelete(SCPICmdWrite):
    """The ``HISTogram:DELete`` command.

    Description:
        - This command deletes the specified histogram.

    Usage:
        - Using the ``.write(value)`` method will send the ``HISTogram:DELete value`` command.

    SCPI Syntax:
        ```
        - HISTogram:DELete <QString>
        ```

    Info:
        - ``<QString>`` specifies the waveform histogram to delete. The argument is of the form '
          HIST<NR1> ', where NR1 is a number value ≥ 1.
    """

    _WRAP_ARG_WITH_QUOTES = True


class HistogramDeleteall(SCPICmdWriteNoArguments):
    """The ``HISTogram:DELETEALL`` command.

    Description:
        - This command deletes all the active instances of histograms defined in the scope
          application.

    Usage:
        - Using the ``.write()`` method will send the ``HISTogram:DELETEALL`` command.

    SCPI Syntax:
        ```
        - HISTogram:DELETEALL
        ```
    """


class HistogramAddnew(SCPICmdWrite):
    """The ``HISTogram:ADDNew`` command.

    Description:
        - This command adds the specified waveform histogram.

    Usage:
        - Using the ``.write(value)`` method will send the ``HISTogram:ADDNew value`` command.

    SCPI Syntax:
        ```
        - HISTogram:ADDNew <QString>
        ```

    Info:
        - ``<QString>`` specifies the waveform histogram to add. The argument is of the form
          'HIST<NR1>', where NR1 is a number value ≥ 1.
    """

    _WRAP_ARG_WITH_QUOTES = True


class Histogram(SCPICmdRead):
    """The ``HISTogram`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``HISTogram?`` query.
        - Using the ``.verify(value)`` method will send the ``HISTogram?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.addnew``: The ``HISTogram:ADDNew`` command.
        - ``.deleteall``: The ``HISTogram:DELETEALL`` command.
        - ``.delete``: The ``HISTogram:DELete`` command.
        - ``.histogram``: The ``HISTogram:HISTogram<x>`` command tree.
        - ``.list``: The ``HISTogram:LIST`` command.
    """

    def __init__(self, device: Optional["PIControl"] = None, cmd_syntax: str = "HISTogram") -> None:
        super().__init__(device, cmd_syntax)
        self._addnew = HistogramAddnew(device, f"{self._cmd_syntax}:ADDNew")
        self._deleteall = HistogramDeleteall(device, f"{self._cmd_syntax}:DELETEALL")
        self._delete = HistogramDelete(device, f"{self._cmd_syntax}:DELete")
        self._histogram: Dict[int, HistogramHistogramItem] = DefaultDictPassKeyToFactory(
            lambda x: HistogramHistogramItem(device, f"{self._cmd_syntax}:HISTogram{x}")
        )
        self._list = HistogramList(device, f"{self._cmd_syntax}:LIST")

    @property
    def addnew(self) -> HistogramAddnew:
        """Return the ``HISTogram:ADDNew`` command.

        Description:
            - This command adds the specified waveform histogram.

        Usage:
            - Using the ``.write(value)`` method will send the ``HISTogram:ADDNew value`` command.

        SCPI Syntax:
            ```
            - HISTogram:ADDNew <QString>
            ```

        Info:
            - ``<QString>`` specifies the waveform histogram to add. The argument is of the form
              'HIST<NR1>', where NR1 is a number value ≥ 1.
        """
        return self._addnew

    @property
    def deleteall(self) -> HistogramDeleteall:
        """Return the ``HISTogram:DELETEALL`` command.

        Description:
            - This command deletes all the active instances of histograms defined in the scope
              application.

        Usage:
            - Using the ``.write()`` method will send the ``HISTogram:DELETEALL`` command.

        SCPI Syntax:
            ```
            - HISTogram:DELETEALL
            ```
        """
        return self._deleteall

    @property
    def delete(self) -> HistogramDelete:
        """Return the ``HISTogram:DELete`` command.

        Description:
            - This command deletes the specified histogram.

        Usage:
            - Using the ``.write(value)`` method will send the ``HISTogram:DELete value`` command.

        SCPI Syntax:
            ```
            - HISTogram:DELete <QString>
            ```

        Info:
            - ``<QString>`` specifies the waveform histogram to delete. The argument is of the form
              ' HIST<NR1> ', where NR1 is a number value ≥ 1.
        """
        return self._delete

    @property
    def histogram(self) -> Dict[int, HistogramHistogramItem]:
        """Return the ``HISTogram:HISTogram<x>`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:HISTogram<x>?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:HISTogram<x>?`` query
              and raise an AssertionError if the returned value does not match ``value``.

        Info:
            - ``HISTogram<x>`` specifies the histogram number.

        Sub-properties:
            - ``.box``: The ``HISTogram:HISTogram<x>:BOX`` command.
            - ``.bstate``: The ``HISTogram:HISTogram<x>:BSTate`` command.
            - ``.data``: The ``HISTogram:HISTogram<x>:DATa`` command.
            - ``.display``: The ``HISTogram:HISTogram<x>:DISPlay`` command.
            - ``.function``: The ``HISTogram:HISTogram<x>:FUNCtion`` command.
            - ``.measurement``: The ``HISTogram:HISTogram<x>:MEASurement`` command tree.
            - ``.save``: The ``HISTogram:HISTogram<x>:SAVe`` command.
            - ``.size``: The ``HISTogram:HISTogram<x>:SIZe`` command.
            - ``.source``: The ``HISTogram:HISTogram<x>:SOUrce`` command.
            - ``.state``: The ``HISTogram:HISTogram<x>:STATE`` command.
            - ``.transparency``: The ``HISTogram:HISTogram<x>:TRANsparency`` command.
        """
        return self._histogram

    @property
    def list(self) -> HistogramList:
        """Return the ``HISTogram:LIST`` command.

        Description:
            - This query returns a comma separated list of all currently defined histograms.

        Usage:
            - Using the ``.query()`` method will send the ``HISTogram:LIST?`` query.
            - Using the ``.verify(value)`` method will send the ``HISTogram:LIST?`` query and raise
              an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - HISTogram:LIST?
            ```
        """
        return self._list



# pyright: reportUnnecessaryTypeIgnoreComment=none
"""Helpers for logging."""

from __future__ import annotations

import importlib.metadata
import inspect
import logging
import sys
import time
import traceback

from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Generator, Literal, Optional, Tuple, Type, TYPE_CHECKING, TypeVar, Union

import colorlog
import pyvisa

from tzlocal import (
    get_localzone,  # pyright: ignore[reportUnknownVariableType]
)

from tm_devices.helpers.enums import CustomStrEnum
from tm_devices.helpers.standalone_helpers import PACKAGE_NAME

if TYPE_CHECKING:
    import os

    from types import TracebackType

_logger_initialized = False


_T = TypeVar("_T", bound=logging.Handler)
__MULTILINE_MESSAGE_LEADING_WHITESPACE = " " * 29
_ORIGINAL_SYS_EXCEPTHOOK = sys.__excepthook__


class _CustomFormatterWithMicroseconds(logging.Formatter):  # pragma: no cover
    """A formatter that includes microsecond precision in the timestamp."""

    converter = datetime.fromtimestamp  # pyright: ignore[reportAssignmentType]

    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None) -> str:  # noqa: N802
        ct: datetime = self.converter(record.created)  # pyright: ignore[reportAssignmentType]
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            # Use microseconds in the time format for 6 digits of precision
            s = f"{t:s}.{ct.microsecond:06d}"
        return s


class _CustomFormatterWithColorAndMicroseconds(
    colorlog.ColoredFormatter, _CustomFormatterWithMicroseconds
):  # pragma: no cover
    """A formatter that includes color and microsecond precision in the timestamp."""


class _CustomFormatterWithPackageNameAndMicroseconds(
    _CustomFormatterWithMicroseconds
):  # pragma: no cover
    """A formatter that includes the package name and microsecond precision in the timestamp."""

    def format(self, record: logging.LogRecord) -> str:
        # Add the package name to the log record
        record.package_name = record.name.split(".", maxsplit=1)[0]
        # Call the original format method
        return super().format(record)


class LoggingLevels(CustomStrEnum):
    """A class holding the valid logging levels supported."""

    DEBUG = "DEBUG"
    """An enum member representing the DEBUG logging level."""
    INFO = "INFO"
    """An enum member representing the INFO logging level."""
    WARNING = "WARNING"
    """An enum member representing the WARNING logging level."""
    ERROR = "ERROR"
    """An enum member representing the ERROR logging level."""
    CRITICAL = "CRITICAL"
    """An enum member representing the CRITICAL logging level."""
    NONE = "NONE"
    """An enum member indicating no logging messages should be captured."""


@contextmanager
def disable_all_loggers(
    level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "CRITICAL",
) -> Generator[None, None, None]:
    """Temporarily disable all logging calls of severity 'level' and below within this context.

    Examples:
        >>> with disable_all_loggers():
        ...     logging.info("This is inside the context and will not be logged")

        >>> logging.info("This is outside the context and will be logged")

        >>> with disable_all_loggers(level="WARNING"):
        ...     logging.info("This will not be logged")
        ...     logging.warning("This will not be logged")
        ...     logging.error("This will be logged since its level is above WARNING")

    Args:
        level: The logging level under which all log messages will be suppressed.
    """
    logging.disable(getattr(logging, level))
    yield
    logging.disable(logging.NOTSET)


def configure_logging(
    *,
    log_console_level: Union[str, LoggingLevels] = LoggingLevels.INFO,
    log_file_level: Union[str, LoggingLevels] = LoggingLevels.DEBUG,
    log_file_directory: Optional[Union[str, os.PathLike[str], Path]] = None,
    log_file_name: Optional[str] = None,
    log_colored_output: bool = False,
    log_pyvisa_messages: bool = False,
    log_uncaught_exceptions: bool = True,
) -> logging.Logger:
    """Configure the logging for this package.

    !!! note
        After this function is called once, if it is called again, it will not perform any
        additional configuration. It will simply return the base logger for the package. This means
        that if logging is configured explicitly in Python code, then any configuration options set
        in the config file or environment variables will be ignored.

    !!! important
        This function will overwrite the [`sys.excepthook`][sys.excepthook] function in order to
        log uncaught exceptions if logging to a log file is enabled. The custom hook function
        used by this package will call [`sys.__excepthook__`][sys.__excepthook__] after the custom
        code is run, so exceptions and tracebacks will still get printed to the console. If you
        have a custom exception hook function you need to use, you will need to overwrite the
        [`sys.excepthook`][sys.excepthook] function after this package's logging is configured.
        To opt out of this behavior and keep Python's default exception handling
        (which means exceptions will not be logged to the log file), set the
        `log_uncaught_exceptions` parameter to `False`.

    Args:
        log_console_level: The logging level to set for the console. Defaults to INFO. Set to
            [`LoggingLevels.NONE`][tm_devices.helpers.logging.LoggingLevels.NONE] to disable all
            console logging/printouts except for certain warnings and exceptions.
        log_file_level: The logging level to set for the file. Defaults to DEBUG. Set to
            [`LoggingLevels.NONE`][tm_devices.helpers.logging.LoggingLevels.NONE] to disable logging
            to a file entirely.
        log_file_directory: The directory to save log files to. Defaults to "./logs" in the
            current working directory.
        log_file_name: The name of the log file to save the logs to. Defaults to a timestamped name
            with the .log extension.
        log_colored_output: Whether to use colored output from the `colorlog` package for the
            console. Defaults to False.
        log_pyvisa_messages: Whether to include logs from the `pyvisa` package in the log file. The
            logging level used for these additional log entries will match
            `log_file_level`. Defaults to False.
        log_uncaught_exceptions: Whether to log uncaught exceptions to the log file with full
            tracebacks and reduce the traceback size of exceptions in the console. Defaults to True.
            Setting the `log_file_level` parameter to
            [`LoggingLevels.NONE`][tm_devices.helpers.logging.LoggingLevels.NONE] will disable
            this feature regardless of the value of `log_uncaught_exceptions`.

    Returns:
        The base logger for the package, this base logger can also be accessed using
            `logging.getLogger(tm_devices.PACKAGE_NAME)`.
    """
    global _logger_initialized  # noqa: PLW0603

    _logger: logging.Logger = logging.getLogger(PACKAGE_NAME)
    if _logger_initialized:
        # If the logger was previously initialized, just return it
        return _logger
    # Convert object types into enum values
    log_console_level = LoggingLevels(log_console_level)
    log_file_level = LoggingLevels(log_file_level)
    # Set the logger level to the lowest level, the handlers will filter out specific levels
    # based on user configuration
    _logger.setLevel(logging.DEBUG)
    _logger.addHandler(logging.NullHandler())
    # The logger/module name is not included in the message, since formatting the messages to
    # be aligned would cause the width of the message prefix to be almost 100 characters before
    # the message is even added to the line.
    logging_file_format_string = "[%(asctime)s] [%(package_name)10s] [%(levelname)8s] %(message)s"
    logging_console_format_string = "%(asctime)s - %(message)s"
    if not log_file_directory:  # pragma: no cover
        log_file_directory = Path("./logs")
    if not log_file_name:  # pragma: no cover
        log_file_name = f"{PACKAGE_NAME}_{time.strftime('%m-%d-%Y_%H-%M-%S', time.localtime())}.log"
    log_filepath = Path(log_file_directory) / log_file_name

    if log_file_level != LoggingLevels.NONE:
        # Set up logger for tm_devices
        log_filepath.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(log_filepath, mode="w", encoding="utf-8")
        file_formatter = _CustomFormatterWithPackageNameAndMicroseconds(logging_file_format_string)

        file_handler.setLevel(getattr(logging, log_file_level.value))
        file_handler.setFormatter(file_formatter)
        _logger.addHandler(file_handler)

        if log_pyvisa_messages:
            # Hook into pyvisa's logging to save it into the same file
            pyvisa.logger.setLevel(getattr(logging, log_file_level.value))
            pyvisa.logger.addHandler(file_handler)

    # Log a few things to just the file
    _logger.debug("timezone==%s", get_localzone())  # pyright: ignore[reportUnknownArgumentType,reportUnnecessaryTypeIgnoreComment]
    _logger.debug("%s==%s", PACKAGE_NAME, importlib.metadata.version(PACKAGE_NAME))

    if log_console_level != LoggingLevels.NONE:
        if log_colored_output:
            console_handler = colorlog.StreamHandler(stream=sys.stdout)
            console_formatter = _CustomFormatterWithColorAndMicroseconds(
                "%(log_color)s" + logging_console_format_string,
                log_colors=colorlog.default_log_colors,
            )
        else:
            console_handler = logging.StreamHandler(stream=sys.stdout)
            console_formatter = _CustomFormatterWithMicroseconds(logging_console_format_string)

        console_handler.setLevel(getattr(logging, log_console_level.value))
        console_handler.setFormatter(console_formatter)
        _logger.addHandler(console_handler)

    if log_uncaught_exceptions and log_file_level != LoggingLevels.NONE:
        sys.excepthook = __exception_handler
    _logger_initialized = True
    return _logger


def __exception_handler(
    exc_type: Type[BaseException], exc_value: BaseException, exc_traceback: TracebackType
) -> None:  # pragma: no cover
    """Log uncaught exceptions."""
    additional_message_for_file = (
        f"\n\nAn uncaught exception occurred. If the exception was explicitly raised by "
        f"the {PACKAGE_NAME} package, look for the most recent previous ERROR log entry "
        f"above to see the specific timestamp of the exception. Otherwise, use the "
        f"traceback to debug the issue."
    )
    updated_message = __log_message_to_console_and_traceback_to_file(
        exc_value.args[0] if exc_value.args else exc_value.__class__.__name__,
        message_for_console="UNCAUGHT EXCEPTION!!!",
        additional_message_for_file=additional_message_for_file,
        exc_info=(exc_type, exc_value, exc_traceback),
    )

    # Remove the original cause if it exists to make the traceback in the console shorter.
    exc_value.__cause__ = None
    # Update the message to include the log file location where the complete traceback is located.
    if exc_value.args:
        exc_value.args = (updated_message, *exc_value.args[1:])
    _ORIGINAL_SYS_EXCEPTHOOK(exc_type, exc_value, exc_traceback)


def __log_to_specific_handler_type_only(
    handler_type: Type[_T],
    message: str,
    logging_level_str: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "EXCEPTION"],
    *,
    exc_info: Optional[Tuple[Type[BaseException], BaseException, TracebackType]] = None,
) -> Optional[_T]:  # pragma: no cover
    """Log a message to handlers of a specific type only.

    This should only be used when absolutely necessary, usually during custom error handling.

    Args:
        handler_type: The type of handler to log the message to.
        message: The message to log.
        logging_level_str: The logging level to log the message at.
        exc_info: The exception information to log. Defaults to None.

    Returns:
        A pointer to the handler that it logged the message to.
    """
    handler_used: Optional[_T] = None
    exception_info: Optional[
        Union[
            Tuple[Type[BaseException], BaseException, TracebackType],
            str,
        ]
    ] = None
    if exc_info:
        exception_info = exc_info
    elif logging_level_str == "EXCEPTION" and sys.exc_info() == (None, None, None):
        exception_info = "Traceback:\n" + "".join(
            traceback.format_stack(inspect.currentframe().f_back.f_back.f_back)  # pyright: ignore[reportOptionalMemberAccess]
        )
    try:
        log_level = getattr(logging, logging_level_str)
    except AttributeError:
        # This is from the EXCEPTION case, so we need to handle it differently
        log_level = logging.CRITICAL
    if handler_type.__name__ == "StreamHandler":
        message = f"\n{__MULTILINE_MESSAGE_LEADING_WHITESPACE}".join(message.splitlines())
    _logger = configure_logging()
    for handler in _logger.handlers:
        if handler.__class__.__name__ == handler_type.__name__ and isinstance(
            handler, handler_type
        ):
            if not handler_used:
                handler_used = handler
            temp_logger = logging.getLogger(f"{__name__}_{handler_type.__name__}_only")
            temp_logger.addHandler(handler)
            temp_logger.setLevel(log_level)
            temp_logger.propagate = False
            if isinstance(exception_info, str):
                temp_logger.log(log_level, "%s%s", message, exception_info)
            elif exc_info:
                temp_logger.log(log_level, message, exc_info=exc_info)
            else:
                temp_logger.log(log_level, message, exc_info=logging_level_str == "EXCEPTION")

    return handler_used


def __log_message_to_console_and_traceback_to_file(
    message: str,
    *,
    message_for_console: Optional[str] = None,
    additional_message_for_file: str = "",
    exc_info: Optional[Tuple[Type[BaseException], BaseException, TracebackType]] = None,
) -> str:  # pyright: ignore[reportUnusedFunction]  # pragma: no cover
    """Log a message to the console and the current traceback to the log file.

    This should only be used when absolutely necessary, usually during custom error handling.

    Args:
        message: The primary message to log.
        message_for_console: The message to log to the console. Defaults to None. Overrides the
            `message` parameter if present.
        additional_message_for_file: An additional message to log to the file. Defaults to None
        exc_info: The exception information to log. Defaults to None.

    Returns:
        The message that was logged, with additional text appended indicating the
            complete traceback is located in the log file.
    """
    file_handler = __log_to_specific_handler_type_only(
        logging.FileHandler,
        message.split("\n")[0] + additional_message_for_file + "\n\n",
        "EXCEPTION",
        exc_info=exc_info,
    )
    __log_to_specific_handler_type_only(
        logging.StreamHandler, message_for_console or message, "CRITICAL"
    )
    if file_handler:
        message += (
            f"\n\nSee the logfile at '{file_handler.baseFilename}' for the complete traceback."
        )
    else:
        message += (
            "No logfile was created, so the complete traceback is not available. "
            "To ensure the complete traceback is available, please update the logging "
            "configuration to include a logfile."
        )
    return message



"""The license commands module.

These commands are used in the following models:
DPO7AX, LPD6, MSO4, MSO4B, MSO5, MSO5B, MSO5LP, MSO6, MSO6B

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:
    ```
    - LICense:APPID? {<QString>}
    - LICense:COUNt?
    - LICense:ERRor?
    - LICense:GMT?
    - LICense:HID?
    - LICense:INSTall <Block>
    - LICense:ITEM? <NR1>
    - LICense:LIST?
    - LICense:UNINSTALL? <QString>
    - LICense:VALidate? <QString>
    - LICense?
    ```
"""

from typing import Optional, TYPE_CHECKING

from ..helpers import SCPICmdRead, SCPICmdReadWithArguments, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.driver_mixins.device_control.pi_control import PIControl


class LicenseValidate(SCPICmdReadWithArguments):
    """The ``LICense:VALidate`` command.

    Description:
        - This query accepts a license nomenclature as an argument and returns True (1) if that
          nomenclature is active and any required hardware is installed, or False (0) if either the
          nomenclature is not active or required hardware is not installed.

    Usage:
        - Using the ``.query(argument)`` method will send the ``LICense:VALidate? argument`` query.
        - Using the ``.verify(argument, value)`` method will send the ``LICense:VALidate? argument``
          query and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:VALidate? <QString>
        ```

    Info:
        - ``<QString>`` is the license nomenclature.
    """


class LicenseUninstall(SCPICmdReadWithArguments):
    """The ``LICense:UNINSTALL`` command.

    Description:
        - Returns the exit license indicated for the user to return to their TekAMS account. Active
          licenses can be specified by their nomenclature. TransactionIDs can be used to specify an
          active license or a previously uninstalled license. In either case, the exit-license is
          returned as block data.

    Usage:
        - Using the ``.query(argument)`` method will send the ``LICense:UNINSTALL? argument`` query.
        - Using the ``.verify(argument, value)`` method will send the
          ``LICense:UNINSTALL? argument`` query and raise an AssertionError if the returned value
          does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:UNINSTALL? <QString>
        ```

    Info:
        - ``<QString>`` is the nomenclature of an active license or a TransactionIDs to specify an
          active license or a previously uninstalled license.
    """


class LicenseList(SCPICmdRead):
    """The ``LICense:LIST`` command.

    Description:
        - This query returns the active license nomenclatures as a comma-separated list of strings.
          Duplicate nomenclatures, that is, the same license but with different expiration dates,
          are included.

    Usage:
        - Using the ``.query()`` method will send the ``LICense:LIST?`` query.
        - Using the ``.verify(value)`` method will send the ``LICense:LIST?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:LIST?
        ```
    """


class LicenseItem(SCPICmdReadWithArguments):
    """The ``LICense:ITEM`` command.

    Description:
        - This query returns the details pertaining to a specific license. The NR1 argument is
          zero-indexed. If no argument is provided, zero is assumed.

    Usage:
        - Using the ``.query(argument)`` method will send the ``LICense:ITEM? argument`` query.
        - Using the ``.verify(argument, value)`` method will send the ``LICense:ITEM? argument``
          query and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:ITEM? <NR1>
        ```

    Info:
        - ``<NR1>`` is the zero-indexed argument specifying a specific license.
    """


class LicenseInstall(SCPICmdWrite):
    """The ``LICense:INSTall`` command.

    Description:
        - This command accepts a block data license and installs it on the instrument. Restarting
          the instrument may be necessary to fully activate the additional capabilities. The format
          of the block data in the argument must have the following: #<File size (Number of
          characters in LIC file)> <LIC file content>

    Usage:
        - Using the ``.write(value)`` method will send the ``LICense:INSTall value`` command.

    SCPI Syntax:
        ```
        - LICense:INSTall <Block>
        ```

    Info:
        - ``<Block>`` is the license in block data format.
    """


class LicenseHid(SCPICmdRead):
    """The ``LICense:HID`` command.

    Description:
        - This query returns the instrument HostID unique identifier.

    Usage:
        - Using the ``.query()`` method will send the ``LICense:HID?`` query.
        - Using the ``.verify(value)`` method will send the ``LICense:HID?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:HID?
        ```
    """


class LicenseGmt(SCPICmdRead):
    """The ``LICense:GMT`` command.

    Description:
        - This query returns the GMT time in ISO 8601 format, the local date, 24 hour time and
          time-zone offset.

    Usage:
        - Using the ``.query()`` method will send the ``LICense:GMT?`` query.
        - Using the ``.verify(value)`` method will send the ``LICense:GMT?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:GMT?
        ```
    """


class LicenseError(SCPICmdRead):
    """The ``LICense:ERRor`` command.

    Description:
        - This query-only command prompts the instrument to return all events and their messages
          (delimited by commas), and removes the returned events from the Event Queue. This command
          is an alias for ALLEV?.

    Usage:
        - Using the ``.query()`` method will send the ``LICense:ERRor?`` query.
        - Using the ``.verify(value)`` method will send the ``LICense:ERRor?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:ERRor?
        ```
    """


class LicenseCount(SCPICmdRead):
    """The ``LICense:COUNt`` command.

    Description:
        - This query returns a count of the number of active licenses installed.

    Usage:
        - Using the ``.query()`` method will send the ``LICense:COUNt?`` query.
        - Using the ``.verify(value)`` method will send the ``LICense:COUNt?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:COUNt?
        ```
    """


class LicenseAppid(SCPICmdReadWithArguments):
    """The ``LICense:APPID`` command.

    Description:
        - This query returns a comma-separated list of the active application IDs. If a string
          argument is provided, a '0' or '1' is returned, according to whether the string matches an
          active application ID.

    Usage:
        - Using the ``.query(argument)`` method will send the ``LICense:APPID? argument`` query.
        - Using the ``.verify(argument, value)`` method will send the ``LICense:APPID? argument``
          query and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense:APPID? {<QString>}
        ```
    """


#  pylint: disable=too-many-instance-attributes
class License(SCPICmdRead):
    """The ``LICense`` command.

    Description:
        - This query-only command returns all license parameters.

    Usage:
        - Using the ``.query()`` method will send the ``LICense?`` query.
        - Using the ``.verify(value)`` method will send the ``LICense?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - LICense?
        ```

    Properties:
        - ``.appid``: The ``LICense:APPID`` command.
        - ``.count``: The ``LICense:COUNt`` command.
        - ``.error``: The ``LICense:ERRor`` command.
        - ``.gmt``: The ``LICense:GMT`` command.
        - ``.hid``: The ``LICense:HID`` command.
        - ``.install``: The ``LICense:INSTall`` command.
        - ``.item``: The ``LICense:ITEM`` command.
        - ``.list``: The ``LICense:LIST`` command.
        - ``.uninstall``: The ``LICense:UNINSTALL`` command.
        - ``.validate``: The ``LICense:VALidate`` command.
    """

    def __init__(self, device: Optional["PIControl"] = None, cmd_syntax: str = "LICense") -> None:
        super().__init__(device, cmd_syntax)
        self._appid = LicenseAppid(device, f"{self._cmd_syntax}:APPID")
        self._count = LicenseCount(device, f"{self._cmd_syntax}:COUNt")
        self._error = LicenseError(device, f"{self._cmd_syntax}:ERRor")
        self._gmt = LicenseGmt(device, f"{self._cmd_syntax}:GMT")
        self._hid = LicenseHid(device, f"{self._cmd_syntax}:HID")
        self._install = LicenseInstall(device, f"{self._cmd_syntax}:INSTall")
        self._item = LicenseItem(device, f"{self._cmd_syntax}:ITEM")
        self._list = LicenseList(device, f"{self._cmd_syntax}:LIST")
        self._uninstall = LicenseUninstall(device, f"{self._cmd_syntax}:UNINSTALL")
        self._validate = LicenseValidate(device, f"{self._cmd_syntax}:VALidate")

    @property
    def appid(self) -> LicenseAppid:
        """Return the ``LICense:APPID`` command.

        Description:
            - This query returns a comma-separated list of the active application IDs. If a string
              argument is provided, a '0' or '1' is returned, according to whether the string
              matches an active application ID.

        Usage:
            - Using the ``.query(argument)`` method will send the ``LICense:APPID? argument`` query.
            - Using the ``.verify(argument, value)`` method will send the
              ``LICense:APPID? argument`` query and raise an AssertionError if the returned value
              does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:APPID? {<QString>}
            ```
        """
        return self._appid

    @property
    def count(self) -> LicenseCount:
        """Return the ``LICense:COUNt`` command.

        Description:
            - This query returns a count of the number of active licenses installed.

        Usage:
            - Using the ``.query()`` method will send the ``LICense:COUNt?`` query.
            - Using the ``.verify(value)`` method will send the ``LICense:COUNt?`` query and raise
              an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:COUNt?
            ```
        """
        return self._count

    @property
    def error(self) -> LicenseError:
        """Return the ``LICense:ERRor`` command.

        Description:
            - This query-only command prompts the instrument to return all events and their messages
              (delimited by commas), and removes the returned events from the Event Queue. This
              command is an alias for ALLEV?.

        Usage:
            - Using the ``.query()`` method will send the ``LICense:ERRor?`` query.
            - Using the ``.verify(value)`` method will send the ``LICense:ERRor?`` query and raise
              an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:ERRor?
            ```
        """
        return self._error

    @property
    def gmt(self) -> LicenseGmt:
        """Return the ``LICense:GMT`` command.

        Description:
            - This query returns the GMT time in ISO 8601 format, the local date, 24 hour time and
              time-zone offset.

        Usage:
            - Using the ``.query()`` method will send the ``LICense:GMT?`` query.
            - Using the ``.verify(value)`` method will send the ``LICense:GMT?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:GMT?
            ```
        """
        return self._gmt

    @property
    def hid(self) -> LicenseHid:
        """Return the ``LICense:HID`` command.

        Description:
            - This query returns the instrument HostID unique identifier.

        Usage:
            - Using the ``.query()`` method will send the ``LICense:HID?`` query.
            - Using the ``.verify(value)`` method will send the ``LICense:HID?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:HID?
            ```
        """
        return self._hid

    @property
    def install(self) -> LicenseInstall:
        """Return the ``LICense:INSTall`` command.

        Description:
            - This command accepts a block data license and installs it on the instrument.
              Restarting the instrument may be necessary to fully activate the additional
              capabilities. The format of the block data in the argument must have the following:
              #<File size (Number of characters in LIC file)> <LIC file content>

        Usage:
            - Using the ``.write(value)`` method will send the ``LICense:INSTall value`` command.

        SCPI Syntax:
            ```
            - LICense:INSTall <Block>
            ```

        Info:
            - ``<Block>`` is the license in block data format.
        """
        return self._install

    @property
    def item(self) -> LicenseItem:
        """Return the ``LICense:ITEM`` command.

        Description:
            - This query returns the details pertaining to a specific license. The NR1 argument is
              zero-indexed. If no argument is provided, zero is assumed.

        Usage:
            - Using the ``.query(argument)`` method will send the ``LICense:ITEM? argument`` query.
            - Using the ``.verify(argument, value)`` method will send the ``LICense:ITEM? argument``
              query and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:ITEM? <NR1>
            ```

        Info:
            - ``<NR1>`` is the zero-indexed argument specifying a specific license.
        """
        return self._item

    @property
    def list(self) -> LicenseList:
        """Return the ``LICense:LIST`` command.

        Description:
            - This query returns the active license nomenclatures as a comma-separated list of
              strings. Duplicate nomenclatures, that is, the same license but with different
              expiration dates, are included.

        Usage:
            - Using the ``.query()`` method will send the ``LICense:LIST?`` query.
            - Using the ``.verify(value)`` method will send the ``LICense:LIST?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:LIST?
            ```
        """
        return self._list

    @property
    def uninstall(self) -> LicenseUninstall:
        """Return the ``LICense:UNINSTALL`` command.

        Description:
            - Returns the exit license indicated for the user to return to their TekAMS account.
              Active licenses can be specified by their nomenclature. TransactionIDs can be used to
              specify an active license or a previously uninstalled license. In either case, the
              exit-license is returned as block data.

        Usage:
            - Using the ``.query(argument)`` method will send the ``LICense:UNINSTALL? argument``
              query.
            - Using the ``.verify(argument, value)`` method will send the
              ``LICense:UNINSTALL? argument`` query and raise an AssertionError if the returned
              value does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:UNINSTALL? <QString>
            ```

        Info:
            - ``<QString>`` is the nomenclature of an active license or a TransactionIDs to specify
              an active license or a previously uninstalled license.
        """
        return self._uninstall

    @property
    def validate(self) -> LicenseValidate:
        """Return the ``LICense:VALidate`` command.

        Description:
            - This query accepts a license nomenclature as an argument and returns True (1) if that
              nomenclature is active and any required hardware is installed, or False (0) if either
              the nomenclature is not active or required hardware is not installed.

        Usage:
            - Using the ``.query(argument)`` method will send the ``LICense:VALidate? argument``
              query.
            - Using the ``.verify(argument, value)`` method will send the
              ``LICense:VALidate? argument`` query and raise an AssertionError if the returned value
              does not match ``value``.

        SCPI Syntax:
            ```
            - LICense:VALidate? <QString>
            ```

        Info:
            - ``<QString>`` is the license nomenclature.
        """
        return self._validate



"""The marker commands module.

These commands are used in the following models:
DPO4K, DPO4KB, MDO3, MDO3K, MDO4K, MDO4KB, MDO4KC, MSO4K, MSO4KB

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:
    ```
    - MARKER:M<x>:AMPLitude:ABSolute?
    - MARKER:M<x>:AMPLitude:DELTa?
    - MARKER:M<x>:FREQuency:ABSolute <NR3>
    - MARKER:M<x>:FREQuency:ABSolute?
    - MARKER:M<x>:FREQuency:DELTa?
    - MARKER:M<x>:NOISEDensity?
    - MARKER:M<x>:PHASENoise?
    - MARKER:MANual {ON|OFF|1|0}
    - MARKER:MANual?
    - MARKER:PEAK:EXCURsion <NR3>
    - MARKER:PEAK:EXCURsion?
    - MARKER:PEAK:MAXimum <NR1>
    - MARKER:PEAK:MAXimum?
    - MARKER:PEAK:STATE {ON|OFF|1|0}
    - MARKER:PEAK:STATE?
    - MARKER:PEAK:THReshold <NR3>
    - MARKER:PEAK:THReshold?
    - MARKER:REFERence CENTER
    - MARKER:REFERence:AMPlitude?
    - MARKER:REFERence:FREQuency?
    - MARKER:TYPe {DELTa|ABSolute}
    - MARKER:TYPe?
    ```
"""

from typing import Dict, Optional, TYPE_CHECKING

from ..helpers import (
    DefaultDictPassKeyToFactory,
    SCPICmdRead,
    SCPICmdWrite,
    ValidatedDynamicNumberCmd,
)

if TYPE_CHECKING:
    from tm_devices.driver_mixins.device_control.pi_control import PIControl


class MarkerType(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:TYPe`` command.

    Description:
        - This command specifies the marker type (either DELTa or ABSolute). An absolute marker
          shows the frequency and amplitude at the location of the marker. A delta marker shows the
          frequency and amplitude of the marker relative to the Reference Marker. The Reference
          Marker shows the absolute frequency and amplitude, regardless of this command. The marker
          amplitude measurements are in dBm for absolute, or in dBc (dB below carrier amplitude) for
          delta.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:TYPe?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:TYPe?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``MARKER:TYPe value`` command.

    SCPI Syntax:
        ```
        - MARKER:TYPe {DELTa|ABSolute}
        - MARKER:TYPe?
        ```

    Info:
        - ``DELTa`` specifies to display the frequency and amplitude of the markers relative to the
          Reference Marker. The relative amplitude is in dBc (dB below carrier amplitude); the
          relative frequency is in Hz.
        - ``ABSolute`` specifies to display the actual frequency and amplitude of each marker. The
          absolute amplitude is in user-set units; the absolute frequency is in Hz.
    """


class MarkerReferenceFrequency(SCPICmdRead):
    """The ``MARKER:REFERence:FREQuency`` command.

    Description:
        - This query returns the frequency of the Reference Marker, in Hz, when the frequency domain
          trace markers are on (using either the command ``MARKER:PEAK:STATE`` or
          ``MARKER:MANUAL``). This data is equivalent to the number that appears on the display next
          to the red R inside a triangle when markers areon. If all markers are off, the value
          returned will be the last value displayed.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:REFERence:FREQuency?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:REFERence:FREQuency?`` query
          and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - MARKER:REFERence:FREQuency?
        ```
    """


class MarkerReferenceAmplitude(SCPICmdRead):
    """The ``MARKER:REFERence:AMPlitude`` command.

    Description:
        - This query returns the actual amplitude (vertical) value of the Reference Marker in
          user-set units. This value indicates the absolute amplitude of the Reference Marker,
          regardless of whether the other markers are manual or automatic. This data is equivalent
          to the number that appears on the display next to the red R inside a triangle when markers
          are turned on. If all markers are turned off, the value returned will be the last value
          displayed.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:REFERence:AMPlitude?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:REFERence:AMPlitude?`` query
          and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - MARKER:REFERence:AMPlitude?
        ```
    """


class MarkerReference(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:REFERence`` command.

    Description:
        - This command changes the Center Frequency to the frequency indicated by the Reference
          Marker, in effect moving the Reference Marker to the center of the screen. This applies
          when markers are turned on (using the command ``MARKER:PEAK:STATE`` or ``MARKER:MANUAL``).
          This is equivalent to the 'R' to Center side menu button in the front panel Markers menu.

    Usage:
        - Using the ``.write(value)`` method will send the ``MARKER:REFERence value`` command.

    SCPI Syntax:
        ```
        - MARKER:REFERence CENTER
        ```

    Properties:
        - ``.amplitude``: The ``MARKER:REFERence:AMPlitude`` command.
        - ``.frequency``: The ``MARKER:REFERence:FREQuency`` command.
    """

    def __init__(self, device: Optional["PIControl"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._amplitude = MarkerReferenceAmplitude(device, f"{self._cmd_syntax}:AMPlitude")
        self._frequency = MarkerReferenceFrequency(device, f"{self._cmd_syntax}:FREQuency")

    @property
    def amplitude(self) -> MarkerReferenceAmplitude:
        """Return the ``MARKER:REFERence:AMPlitude`` command.

        Description:
            - This query returns the actual amplitude (vertical) value of the Reference Marker in
              user-set units. This value indicates the absolute amplitude of the Reference Marker,
              regardless of whether the other markers are manual or automatic. This data is
              equivalent to the number that appears on the display next to the red R inside a
              triangle when markers are turned on. If all markers are turned off, the value returned
              will be the last value displayed.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:REFERence:AMPlitude?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:REFERence:AMPlitude?``
              query and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - MARKER:REFERence:AMPlitude?
            ```
        """
        return self._amplitude

    @property
    def frequency(self) -> MarkerReferenceFrequency:
        """Return the ``MARKER:REFERence:FREQuency`` command.

        Description:
            - This query returns the frequency of the Reference Marker, in Hz, when the frequency
              domain trace markers are on (using either the command ``MARKER:PEAK:STATE`` or
              ``MARKER:MANUAL``). This data is equivalent to the number that appears on the display
              next to the red R inside a triangle when markers areon. If all markers are off, the
              value returned will be the last value displayed.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:REFERence:FREQuency?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:REFERence:FREQuency?``
              query and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - MARKER:REFERence:FREQuency?
            ```
        """
        return self._frequency


class MarkerPeakThreshold(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:PEAK:THReshold`` command.

    Description:
        - This command specifies the threshold value of the automatic peak markers available for
          frequency domain traces. (Use the ``RF:UNITS`` command to specify the units.) Only peaks
          with an amplitude greater than the threshold value will qualify for automatic peak marker
          placement. To set the excursion value for the automatic markers, use the command
          ``MARKER:PEAK:EXCURSION``. To switch the automatic marker readout between absolute and
          delta, use the command ``MARKER:TYPE``. To turn on and set the number of automatic
          markers, use the commands ``MARKER:PEAK:STATE`` and ``MARKER:PEAK:MAXIMUM``. To list all
          of the peak markers, use the command ``SEARCH:SPECTRAL:LIST``

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:PEAK:THReshold?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:THReshold?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``MARKER:PEAK:THReshold value`` command.

    SCPI Syntax:
        ```
        - MARKER:PEAK:THReshold <NR3>
        - MARKER:PEAK:THReshold?
        ```

    Info:
        - ``<NR3>`` is a floating point number that indicates the automatic marker threshold value.
    """


class MarkerPeakState(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:PEAK:STATE`` command.

    Description:
        - This command switches on or off the automatic peak markers that are available for
          frequency domain traces. There are up to 11 automatic markers. The maximum number of
          markers can be set using the command ``MARKER:PEAK:MAXIMUM``. The automatic peak markers
          find amplitude peaks based upon user threshold and excursion settings (set with the
          ``MARKER:PEAK:EXCURSION`` and ``MARKER:PEAK:THRESHOLD`` commands.) Each automatic marker
          has a readout associated with it. These can be absolute or delta readouts (set with the
          ``MARKER:TYPE`` command.) To list all of the peak markers, use the command
          ``SEARCH:SPECTRAL:LIST``

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:PEAK:STATE?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:STATE?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``MARKER:PEAK:STATE value`` command.

    SCPI Syntax:
        ```
        - MARKER:PEAK:STATE {ON|OFF|1|0}
        - MARKER:PEAK:STATE?
        ```

    Info:
        - ``OFF`` or 0 turns the automatic peak markers off.
        - ``ON`` or 1 turns the automatic peak markers on.
    """


class MarkerPeakMaximum(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:PEAK:MAXimum`` command.

    Description:
        - This command specifies the maximum number of frequency domain trace peaks that could have
          automatic markers placed on them. This can be a number between 1 and 11. The default is 5.
          To turn on the automatic peak markers, use the command ``MARKER:PEAK:STATE``. To list all
          of the peak markers, use the command ``SEARCH:SPECTRAL:LIST The`` actual number of
          automatic markers may be less than the maximum, depending on the threshold and excursion
          values and the spectral content of the RF signal. If more peaks than the maximum are
          detected that meet the threshold and excursion criteria, only the highest amplitude peaks
          will have automatic markers placed on them.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:PEAK:MAXimum?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:MAXimum?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``MARKER:PEAK:MAXimum value`` command.

    SCPI Syntax:
        ```
        - MARKER:PEAK:MAXimum <NR1>
        - MARKER:PEAK:MAXimum?
        ```

    Info:
        - ``<NR1>`` is an integer that represents the maximum number of peaks that could have
          automatic markers.
    """


class MarkerPeakExcursion(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:PEAK:EXCURsion`` command.

    Description:
        - This command specifies the peak excursion value, in dB, for the frequency domain trace
          automatic peak markers. Peak excursion refers to how far an RF signal needs to fall in
          amplitude between marked peaks, in order to be considered another valid peak. If the peak
          excursion value is low, more peaks will tend to qualify as valid peaks and have associated
          markers. If the peak excursion value is high, fewer peaks will tend to qualify as valid
          peaks and have associated markers.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:PEAK:EXCURsion?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:EXCURsion?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``MARKER:PEAK:EXCURsion value`` command.

    SCPI Syntax:
        ```
        - MARKER:PEAK:EXCURsion <NR3>
        - MARKER:PEAK:EXCURsion?
        ```

    Info:
        - ``<NR3>`` is a floating point number that specifies the automatic marker excursion value.
    """


class MarkerPeak(SCPICmdRead):
    """The ``MARKER:PEAK`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:PEAK?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:PEAK?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.excursion``: The ``MARKER:PEAK:EXCURsion`` command.
        - ``.maximum``: The ``MARKER:PEAK:MAXimum`` command.
        - ``.state``: The ``MARKER:PEAK:STATE`` command.
        - ``.threshold``: The ``MARKER:PEAK:THReshold`` command.
    """

    def __init__(self, device: Optional["PIControl"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._excursion = MarkerPeakExcursion(device, f"{self._cmd_syntax}:EXCURsion")
        self._maximum = MarkerPeakMaximum(device, f"{self._cmd_syntax}:MAXimum")
        self._state = MarkerPeakState(device, f"{self._cmd_syntax}:STATE")
        self._threshold = MarkerPeakThreshold(device, f"{self._cmd_syntax}:THReshold")

    @property
    def excursion(self) -> MarkerPeakExcursion:
        """Return the ``MARKER:PEAK:EXCURsion`` command.

        Description:
            - This command specifies the peak excursion value, in dB, for the frequency domain trace
              automatic peak markers. Peak excursion refers to how far an RF signal needs to fall in
              amplitude between marked peaks, in order to be considered another valid peak. If the
              peak excursion value is low, more peaks will tend to qualify as valid peaks and have
              associated markers. If the peak excursion value is high, fewer peaks will tend to
              qualify as valid peaks and have associated markers.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:PEAK:EXCURsion?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:EXCURsion?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``MARKER:PEAK:EXCURsion value``
              command.

        SCPI Syntax:
            ```
            - MARKER:PEAK:EXCURsion <NR3>
            - MARKER:PEAK:EXCURsion?
            ```

        Info:
            - ``<NR3>`` is a floating point number that specifies the automatic marker excursion
              value.
        """
        return self._excursion

    @property
    def maximum(self) -> MarkerPeakMaximum:
        """Return the ``MARKER:PEAK:MAXimum`` command.

        Description:
            - This command specifies the maximum number of frequency domain trace peaks that could
              have automatic markers placed on them. This can be a number between 1 and 11. The
              default is 5. To turn on the automatic peak markers, use the command
              ``MARKER:PEAK:STATE``. To list all of the peak markers, use the command
              ``SEARCH:SPECTRAL:LIST The`` actual number of automatic markers may be less than the
              maximum, depending on the threshold and excursion values and the spectral content of
              the RF signal. If more peaks than the maximum are detected that meet the threshold and
              excursion criteria, only the highest amplitude peaks will have automatic markers
              placed on them.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:PEAK:MAXimum?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:MAXimum?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``MARKER:PEAK:MAXimum value``
              command.

        SCPI Syntax:
            ```
            - MARKER:PEAK:MAXimum <NR1>
            - MARKER:PEAK:MAXimum?
            ```

        Info:
            - ``<NR1>`` is an integer that represents the maximum number of peaks that could have
              automatic markers.
        """
        return self._maximum

    @property
    def state(self) -> MarkerPeakState:
        """Return the ``MARKER:PEAK:STATE`` command.

        Description:
            - This command switches on or off the automatic peak markers that are available for
              frequency domain traces. There are up to 11 automatic markers. The maximum number of
              markers can be set using the command ``MARKER:PEAK:MAXIMUM``. The automatic peak
              markers find amplitude peaks based upon user threshold and excursion settings (set
              with the ``MARKER:PEAK:EXCURSION`` and ``MARKER:PEAK:THRESHOLD`` commands.) Each
              automatic marker has a readout associated with it. These can be absolute or delta
              readouts (set with the ``MARKER:TYPE`` command.) To list all of the peak markers, use
              the command ``SEARCH:SPECTRAL:LIST``

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:PEAK:STATE?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:STATE?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``MARKER:PEAK:STATE value`` command.

        SCPI Syntax:
            ```
            - MARKER:PEAK:STATE {ON|OFF|1|0}
            - MARKER:PEAK:STATE?
            ```

        Info:
            - ``OFF`` or 0 turns the automatic peak markers off.
            - ``ON`` or 1 turns the automatic peak markers on.
        """
        return self._state

    @property
    def threshold(self) -> MarkerPeakThreshold:
        """Return the ``MARKER:PEAK:THReshold`` command.

        Description:
            - This command specifies the threshold value of the automatic peak markers available for
              frequency domain traces. (Use the ``RF:UNITS`` command to specify the units.) Only
              peaks with an amplitude greater than the threshold value will qualify for automatic
              peak marker placement. To set the excursion value for the automatic markers, use the
              command ``MARKER:PEAK:EXCURSION``. To switch the automatic marker readout between
              absolute and delta, use the command ``MARKER:TYPE``. To turn on and set the number of
              automatic markers, use the commands ``MARKER:PEAK:STATE`` and ``MARKER:PEAK:MAXIMUM``.
              To list all of the peak markers, use the command ``SEARCH:SPECTRAL:LIST``

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:PEAK:THReshold?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:PEAK:THReshold?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``MARKER:PEAK:THReshold value``
              command.

        SCPI Syntax:
            ```
            - MARKER:PEAK:THReshold <NR3>
            - MARKER:PEAK:THReshold?
            ```

        Info:
            - ``<NR3>`` is a floating point number that indicates the automatic marker threshold
              value.
        """
        return self._threshold


class MarkerManual(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:MANual`` command.

    Description:
        - This command switches on or off the manual markers a and b that are available for
          frequency domain traces. Two manual markers are available for measuring non-peak areas of
          interest. The absolute measurements are in dBm; the relative measurements (relative to the
          Reference Marker) are in dBc (dB relative to the carrier). When the manual markers are
          turned off, and the peak markers are turned on, the Reference Marker is placed on the
          highest amplitude peak. With manual markers on, the Reference Marker is placed at the a
          manual marker. The manual markers use the units specified with the command ``RF:UNITS``.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:MANual?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:MANual?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``MARKER:MANual value`` command.

    SCPI Syntax:
        ```
        - MARKER:MANual {ON|OFF|1|0}
        - MARKER:MANual?
        ```

    Info:
        - ``OFF`` or 0 turns the manual markers off.
        - ``ON`` or 1 turns the manual markers on.
    """


class MarkerMItemPhasenoise(SCPICmdRead):
    """The ``MARKER:M<x>:PHASENoise`` command.

    Description:
        - This command returns the phase noise of the ``RF_NORMal`` trace at the specified marker
          position in dBc/Hz units.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:PHASENoise?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:PHASENoise?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - MARKER:M<x>:PHASENoise?
        ```
    """


class MarkerMItemNoisedensity(SCPICmdRead):
    """The ``MARKER:M<x>:NOISEDensity`` command.

    Description:
        - This command returns the noise density of the ``RF_NORMal`` trace at the specified marker
          position in <RF Units>/Hz units, where <RF Units> are the units specified by the command
          ``RF:UNITS``.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:NOISEDensity?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:NOISEDensity?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - MARKER:M<x>:NOISEDensity?
        ```
    """


class MarkerMItemFrequencyDelta(SCPICmdRead):
    """The ``MARKER:M<x>:FREQuency:DELTa`` command.

    Description:
        - This query returns the delta frequency (horizontal) value of either of the two manual
          markers that are available for frequency domain traces, in relation to the Reference
          Marker. M<x> can be either M1, which specifies manual marker a, or M2, which specifies
          manual marker b. The manual marker readouts use the units specified with the command
          ``RF:UNITS``.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:FREQuency:DELTa?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:FREQuency:DELTa?`` query
          and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - MARKER:M<x>:FREQuency:DELTa?
        ```
    """


class MarkerMItemFrequencyAbsolute(SCPICmdWrite, SCPICmdRead):
    """The ``MARKER:M<x>:FREQuency:ABSolute`` command.

    Description:
        - This command specifies the actual frequency (horizontal) value of either of the two manual
          markers that are available for frequency domain traces. M<x> can be either M1, which
          specifies manual marker a, or M2, which specifies manual marker b.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:FREQuency:ABSolute?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:FREQuency:ABSolute?``
          query and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``MARKER:M<x>:FREQuency:ABSolute value``
          command.

    SCPI Syntax:
        ```
        - MARKER:M<x>:FREQuency:ABSolute <NR3>
        - MARKER:M<x>:FREQuency:ABSolute?
        ```

    Info:
        - ``<NR3>`` is a floating point value that indicates the actual frequency of either of the
          two manual markers.
    """


class MarkerMItemFrequency(SCPICmdRead):
    """The ``MARKER:M<x>:FREQuency`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:FREQuency?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:FREQuency?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.absolute``: The ``MARKER:M<x>:FREQuency:ABSolute`` command.
        - ``.delta``: The ``MARKER:M<x>:FREQuency:DELTa`` command.
    """

    def __init__(self, device: Optional["PIControl"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._absolute = MarkerMItemFrequencyAbsolute(device, f"{self._cmd_syntax}:ABSolute")
        self._delta = MarkerMItemFrequencyDelta(device, f"{self._cmd_syntax}:DELTa")

    @property
    def absolute(self) -> MarkerMItemFrequencyAbsolute:
        """Return the ``MARKER:M<x>:FREQuency:ABSolute`` command.

        Description:
            - This command specifies the actual frequency (horizontal) value of either of the two
              manual markers that are available for frequency domain traces. M<x> can be either M1,
              which specifies manual marker a, or M2, which specifies manual marker b.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:FREQuency:ABSolute?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:FREQuency:ABSolute?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``MARKER:M<x>:FREQuency:ABSolute value`` command.

        SCPI Syntax:
            ```
            - MARKER:M<x>:FREQuency:ABSolute <NR3>
            - MARKER:M<x>:FREQuency:ABSolute?
            ```

        Info:
            - ``<NR3>`` is a floating point value that indicates the actual frequency of either of
              the two manual markers.
        """
        return self._absolute

    @property
    def delta(self) -> MarkerMItemFrequencyDelta:
        """Return the ``MARKER:M<x>:FREQuency:DELTa`` command.

        Description:
            - This query returns the delta frequency (horizontal) value of either of the two manual
              markers that are available for frequency domain traces, in relation to the Reference
              Marker. M<x> can be either M1, which specifies manual marker a, or M2, which specifies
              manual marker b. The manual marker readouts use the units specified with the command
              ``RF:UNITS``.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:FREQuency:DELTa?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:FREQuency:DELTa?``
              query and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - MARKER:M<x>:FREQuency:DELTa?
            ```
        """
        return self._delta


class MarkerMItemAmplitudeDelta(SCPICmdRead):
    """The ``MARKER:M<x>:AMPLitude:DELTa`` command.

    Description:
        - This query returns the delta amplitude (vertical) value of either of the two manual
          markers that are available for frequency domain traces, in relation to the Reference
          Marker. M<x> can be either M1, which specifies manual marker a, or M2, which specifies
          manual marker b.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:AMPLitude:DELTa?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:AMPLitude:DELTa?`` query
          and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - MARKER:M<x>:AMPLitude:DELTa?
        ```
    """


class MarkerMItemAmplitudeAbsolute(SCPICmdRead):
    """The ``MARKER:M<x>:AMPLitude:ABSolute`` command.

    Description:
        - This query returns the actual amplitude (vertical) value of either of the two manual
          markers that are available for frequency domain traces, in user-set units. M<x> can be
          either M1, which specifies manual marker a, or M2, which specifies manual marker b. (Use
          ``RF:UNITS`` to specify the units.)

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:AMPLitude:ABSolute?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:AMPLitude:ABSolute?``
          query and raise an AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - MARKER:M<x>:AMPLitude:ABSolute?
        ```
    """


class MarkerMItemAmplitude(SCPICmdRead):
    """The ``MARKER:M<x>:AMPLitude`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>:AMPLitude?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:AMPLitude?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.absolute``: The ``MARKER:M<x>:AMPLitude:ABSolute`` command.
        - ``.delta``: The ``MARKER:M<x>:AMPLitude:DELTa`` command.
    """

    def __init__(self, device: Optional["PIControl"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._absolute = MarkerMItemAmplitudeAbsolute(device, f"{self._cmd_syntax}:ABSolute")
        self._delta = MarkerMItemAmplitudeDelta(device, f"{self._cmd_syntax}:DELTa")

    @property
    def absolute(self) -> MarkerMItemAmplitudeAbsolute:
        """Return the ``MARKER:M<x>:AMPLitude:ABSolute`` command.

        Description:
            - This query returns the actual amplitude (vertical) value of either of the two manual
              markers that are available for frequency domain traces, in user-set units. M<x> can be
              either M1, which specifies manual marker a, or M2, which specifies manual marker b.
              (Use ``RF:UNITS`` to specify the units.)

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:AMPLitude:ABSolute?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:AMPLitude:ABSolute?``
              query and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - MARKER:M<x>:AMPLitude:ABSolute?
            ```
        """
        return self._absolute

    @property
    def delta(self) -> MarkerMItemAmplitudeDelta:
        """Return the ``MARKER:M<x>:AMPLitude:DELTa`` command.

        Description:
            - This query returns the delta amplitude (vertical) value of either of the two manual
              markers that are available for frequency domain traces, in relation to the Reference
              Marker. M<x> can be either M1, which specifies manual marker a, or M2, which specifies
              manual marker b.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:AMPLitude:DELTa?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:AMPLitude:DELTa?``
              query and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - MARKER:M<x>:AMPLitude:DELTa?
            ```
        """
        return self._delta


class MarkerMItem(ValidatedDynamicNumberCmd, SCPICmdRead):
    """The ``MARKER:M<x>`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER:M<x>?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER:M<x>?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.amplitude``: The ``MARKER:M<x>:AMPLitude`` command tree.
        - ``.frequency``: The ``MARKER:M<x>:FREQuency`` command tree.
        - ``.noisedensity``: The ``MARKER:M<x>:NOISEDensity`` command.
        - ``.phasenoise``: The ``MARKER:M<x>:PHASENoise`` command.
    """

    def __init__(self, device: Optional["PIControl"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._amplitude = MarkerMItemAmplitude(device, f"{self._cmd_syntax}:AMPLitude")
        self._frequency = MarkerMItemFrequency(device, f"{self._cmd_syntax}:FREQuency")
        self._noisedensity = MarkerMItemNoisedensity(device, f"{self._cmd_syntax}:NOISEDensity")
        self._phasenoise = MarkerMItemPhasenoise(device, f"{self._cmd_syntax}:PHASENoise")

    @property
    def amplitude(self) -> MarkerMItemAmplitude:
        """Return the ``MARKER:M<x>:AMPLitude`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:AMPLitude?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:AMPLitude?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.absolute``: The ``MARKER:M<x>:AMPLitude:ABSolute`` command.
            - ``.delta``: The ``MARKER:M<x>:AMPLitude:DELTa`` command.
        """
        return self._amplitude

    @property
    def frequency(self) -> MarkerMItemFrequency:
        """Return the ``MARKER:M<x>:FREQuency`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:FREQuency?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:FREQuency?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.absolute``: The ``MARKER:M<x>:FREQuency:ABSolute`` command.
            - ``.delta``: The ``MARKER:M<x>:FREQuency:DELTa`` command.
        """
        return self._frequency

    @property
    def noisedensity(self) -> MarkerMItemNoisedensity:
        """Return the ``MARKER:M<x>:NOISEDensity`` command.

        Description:
            - This command returns the noise density of the ``RF_NORMal`` trace at the specified
              marker position in <RF Units>/Hz units, where <RF Units> are the units specified by
              the command ``RF:UNITS``.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:NOISEDensity?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:NOISEDensity?`` query
              and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - MARKER:M<x>:NOISEDensity?
            ```
        """
        return self._noisedensity

    @property
    def phasenoise(self) -> MarkerMItemPhasenoise:
        """Return the ``MARKER:M<x>:PHASENoise`` command.

        Description:
            - This command returns the phase noise of the ``RF_NORMal`` trace at the specified
              marker position in dBc/Hz units.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>:PHASENoise?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>:PHASENoise?`` query
              and raise an AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - MARKER:M<x>:PHASENoise?
            ```
        """
        return self._phasenoise


class Marker(SCPICmdRead):
    """The ``MARKER`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``MARKER?`` query.
        - Using the ``.verify(value)`` method will send the ``MARKER?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.m``: The ``MARKER:M<x>`` command tree.
        - ``.manual``: The ``MARKER:MANual`` command.
        - ``.peak``: The ``MARKER:PEAK`` command tree.
        - ``.reference``: The ``MARKER:REFERence`` command.
        - ``.type``: The ``MARKER:TYPe`` command.
    """

    def __init__(self, device: Optional["PIControl"] = None, cmd_syntax: str = "MARKER") -> None:
        super().__init__(device, cmd_syntax)
        self._m: Dict[int, MarkerMItem] = DefaultDictPassKeyToFactory(
            lambda x: MarkerMItem(device, f"{self._cmd_syntax}:M{x}")
        )
        self._manual = MarkerManual(device, f"{self._cmd_syntax}:MANual")
        self._peak = MarkerPeak(device, f"{self._cmd_syntax}:PEAK")
        self._reference = MarkerReference(device, f"{self._cmd_syntax}:REFERence")
        self._type = MarkerType(device, f"{self._cmd_syntax}:TYPe")

    @property
    def m(self) -> Dict[int, MarkerMItem]:
        """Return the ``MARKER:M<x>`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:M<x>?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:M<x>?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.amplitude``: The ``MARKER:M<x>:AMPLitude`` command tree.
            - ``.frequency``: The ``MARKER:M<x>:FREQuency`` command tree.
            - ``.noisedensity``: The ``MARKER:M<x>:NOISEDensity`` command.
            - ``.phasenoise``: The ``MARKER:M<x>:PHASENoise`` command.
        """
        return self._m

    @property
    def manual(self) -> MarkerManual:
        """Return the ``MARKER:MANual`` command.

        Description:
            - This command switches on or off the manual markers a and b that are available for
              frequency domain traces. Two manual markers are available for measuring non-peak areas
              of interest. The absolute measurements are in dBm; the relative measurements (relative
              to the Reference Marker) are in dBc (dB relative to the carrier). When the manual
              markers are turned off, and the peak markers are turned on, the Reference Marker is
              placed on the highest amplitude peak. With manual markers on, the Reference Marker is
              placed at the a manual marker. The manual markers use the units specified with the
              command ``RF:UNITS``.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:MANual?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:MANual?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``MARKER:MANual value`` command.

        SCPI Syntax:
            ```
            - MARKER:MANual {ON|OFF|1|0}
            - MARKER:MANual?
            ```

        Info:
            - ``OFF`` or 0 turns the manual markers off.
            - ``ON`` or 1 turns the manual markers on.
        """
        return self._manual

    @property
    def peak(self) -> MarkerPeak:
        """Return the ``MARKER:PEAK`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:PEAK?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:PEAK?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.excursion``: The ``MARKER:PEAK:EXCURsion`` command.
            - ``.maximum``: The ``MARKER:PEAK:MAXimum`` command.
            - ``.state``: The ``MARKER:PEAK:STATE`` command.
            - ``.threshold``: The ``MARKER:PEAK:THReshold`` command.
        """
        return self._peak

    @property
    def reference(self) -> MarkerReference:
        """Return the ``MARKER:REFERence`` command.

        Description:
            - This command changes the Center Frequency to the frequency indicated by the Reference
              Marker, in effect moving the Reference Marker to the center of the screen. This
              applies when markers are turned on (using the command ``MARKER:PEAK:STATE`` or
              ``MARKER:MANUAL``). This is equivalent to the 'R' to Center side menu button in the
              front panel Markers menu.

        Usage:
            - Using the ``.write(value)`` method will send the ``MARKER:REFERence value`` command.

        SCPI Syntax:
            ```
            - MARKER:REFERence CENTER
            ```

        Sub-properties:
            - ``.amplitude``: The ``MARKER:REFERence:AMPlitude`` command.
            - ``.frequency``: The ``MARKER:REFERence:FREQuency`` command.
        """
        return self._reference

    @property
    def type(self) -> MarkerType:
        """Return the ``MARKER:TYPe`` command.

        Description:
            - This command specifies the marker type (either DELTa or ABSolute). An absolute marker
              shows the frequency and amplitude at the location of the marker. A delta marker shows
              the frequency and amplitude of the marker relative to the Reference Marker. The
              Reference Marker shows the absolute frequency and amplitude, regardless of this
              command. The marker amplitude measurements are in dBm for absolute, or in dBc (dB
              below carrier amplitude) for delta.

        Usage:
            - Using the ``.query()`` method will send the ``MARKER:TYPe?`` query.
            - Using the ``.verify(value)`` method will send the ``MARKER:TYPe?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``MARKER:TYPe value`` command.

        SCPI Syntax:
            ```
            - MARKER:TYPe {DELTa|ABSolute}
            - MARKER:TYPe?
            ```

        Info:
            - ``DELTa`` specifies to display the frequency and amplitude of the markers relative to
              the Reference Marker. The relative amplitude is in dBc (dB below carrier amplitude);
              the relative frequency is in Hz.
            - ``ABSolute`` specifies to display the actual frequency and amplitude of each marker.
              The absolute amplitude is in user-set units; the absolute frequency is in Hz.
        """
        return self._type
